(* WinTouch - a Windows clone of Linux touch
   ========
  calling: wintouch [options] filename1 [filename2 [filename3 ..]]
  command line options
   -cwa  : select timestamp(s) to be modified (default = w)
   -r    : copy timestamps from filename1 (default only w) to filename2,...
     -b:n  : subtract n seconds from time stamp
     -f:n  : add n seconds from time stamp
   -d:[[YYYY]MM]DD : set to fix date, time and omitted parts remain unchanged
   -t:hh[mm[ss]]   : set to fix time, date and omitted parts remain unchanged
   <none of r,d,t> : set all files to current date and time

   create the specified file if not exist

    J. Rathlev, Dr. J. Rathlev, D-24222 Schwentinental

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - Jan. 2016
   last modified: July 2020
   *)

program WinTouch;

{$APPTYPE CONSOLE}

{$R *.res}

uses
  System.SysUtils, System.Classes, Winapi.Windows, System.DateUtils,
  XlFileUtils, FileUtils, StringUtils;

type
  TModTimeStamp = (mtCreate,mtWrite,mtAccess);
  TModTimeStamps = set of TModTimeStamp;

var
  FileList    : TStringList;
  Delta,i,
  nc,nn       : integer;
  HasOptions,
  UseCurrent,
  CopyTs      : boolean;
  ModTs       : TModTimeStamps;
  sDate,sTime : string;
  fts         : TFileTimestamps;

{ ---------------------------------------------------------------- }
function GetVersionInfo : string;
var
  s : string;
  dt : TDateTime;
begin
  dt:=GetFileLastWriteDateTime(ParamStr(0));
  if dt>0 then s:=DateToStr(dt) else s:='?';
  Result:=Format('WinTouch - Version 2.0 (%s)  2016-2020 Dr. J. Rathlev',[s]);
  end;

{ ---------------------------------------------------------------- }
function ReplaceDate (const s : string; dt : TDateTime) : TDateTime;
var
  y,m,d : integer;
  ok    : boolean;
begin
  y:=YearOf(dt); m:=MonthOf(dt);
  if length(s)=0 then ok:=false
  else begin
    if length(s)<=2 then begin // only day
      ok:=TryStrToInt(copy(s,1,2),d);
      end
    else if length(s)<=4 then begin  // month and day
      ok:=TryStrToInt(copy(s,1,2),m) and TryStrToInt(copy(s,3,2),d);
      end
    else begin // year, month and day
      ok:=TryStrToInt(copy(s,1,4),y) and TryStrToInt(copy(s,5,2),m) and TryStrToInt(copy(s,7,2),d);
      end;
    ok:=ok and TryEncodeDate(y,m,d,Result);
    if ok then Result:=Result+TimeOf(dt);
    end;
  if not ok then Result:=dt;
  end;

function ReplaceTime (const s : string; dt : TDateTime) : TDateTime;
var
  h,m,ss : integer;
  ok    : boolean;
begin
  m:=MinuteOf(dt); ss:=SecondOf(dt);
  if length(s)=0 then ok:=false
  else begin
    if length(s)<=2 then begin // only hour
      ok:=TryStrToInt(copy(s,1,2),h);
      end
    else if length(s)<=4 then begin  // hour and minute
      ok:=TryStrToInt(copy(s,1,2),h) and TryStrToInt(copy(s,3,2),m);
      end
    else begin // year, month and day
      ok:=TryStrToInt(copy(s,1,2),h) and TryStrToInt(copy(s,3,2),m) and TryStrToInt(copy(s,5,2),ss);
      end;
    ok:=ok and TryEncodeTime(h,m,ss,0,Result);
    if ok then Result:=Result+DateOf(dt);
    end;
  if not ok then Result:=dt;
  end;

{ ---------------------------------------------------------------- }
function GetCurrentTimestamps : TFileTimestamps;
var
  dt : TDateTime;
begin
  dt:=Now;
  with Result do begin
    LastWriteTime:=LocalDateTimeToFileTime(dt);
    CreationTime:=LocalDateTimeToFileTime(dt);
    LastAccessTime:=LocalDateTimeToFileTime(dt);
    end;
  end;

function SetFileTimestamps (const FileName: string; Timestamps : TFileTimestamps) : boolean;
var
  Handle   : THandle;
  wt,ct,at : PFileTime;
begin
  Result:=true;
  Handle:=CreateFile(PChar(FileNameToXL(FileName)),FILE_WRITE_ATTRIBUTES,0,nil,OPEN_EXISTING,FILE_FLAG_BACKUP_SEMANTICS,0);
  if Handle=INVALID_HANDLE_VALUE then begin
    writeln('Error opening file: ',FileName);
    writeln('  System error:',SysErrorMessage(GetLastError));
    Result:=false;
    end
  else with Timestamps do begin
    if mtWrite in ModTs then wt:=@LastWriteTime else wt:=nil;
    if mtCreate in ModTs then ct:=@CreationTime else ct:=nil;
    if mtAccess in ModTs then at:=@LastAccessTime else at:=nil;
    if not SetFileTime(Handle,ct,at,wt) then begin
      writeln('Error writing timestamps: ',FileName);
      writeln('  System error:',SysErrorMessage(GetLastError));
      Result:=false;
      end;
    FileClose(Handle);
    end;
  end;

//function FileTimeToDateTime (ft : TFileTime) : TDateTime;
//var
//  st : TSystemTime;
//begin
//  if not (FileTimeToSystemTime(ft,st) and TrySystemTimeToDateTime(st,Result)) then
//    Result:=EncodeDate(1899,12,30);
//  end;

procedure ModifyTimestamps (var Timestamps : TFileTimestamps);
var
  dt : TDateTime;
begin
  with TimeStamps do begin
    if mtWrite in ModTs then begin
      dt:=ReplaceTime(sTime,ReplaceDate(sDate,FileTimeToLocalDateTime(LastWriteTime)));
      LastWriteTime:=LocalDateTimeToFileTime(dt);
      end;
    if mtCreate in ModTs then begin
      dt:=ReplaceTime(sTime,ReplaceDate(sDate,FileTimeToLocalDateTime(CreationTime)));
      CreationTime:=LocalDateTimeToFileTime(dt);
      end;
    if mtAccess in ModTs then begin
      dt:=ReplaceTime(sTime,ReplaceDate(sDate,FileTimeToLocalDateTime(LastAccessTime)));
      LastAccessTime:=LocalDateTimeToFileTime(dt);
      end;
    end;
  end;

procedure IncTimeStamps (var Timestamps : TFileTimestamps);
begin
  with TimeStamps do begin
    if mtWrite in ModTs then begin
      LastWriteTime:=DateTimeToFileTime(IncSecond(FileTimeToDateTime(LastWriteTime),Delta));
      end;
    if mtCreate in ModTs then begin
      CreationTime:=DateTimeToFileTime(IncSecond(FileTimeToDateTime(CreationTime),Delta));
      end;
    if mtAccess in ModTs then begin
      LastAccessTime:=DateTimeToFileTime(IncSecond(FileTimeToDateTime(LastAccessTime),Delta));
      end;
    end;
  end;

function CheckFile (const Filename : string; var CCnt : integer) : boolean;
var
  fh : THandle;
begin
  Result:=FileExists(Filename);
  if not Result then begin
    fh:=FileCreate(FileNameToXL(Filename));   // create
    if fh=INVALID_HANDLE_VALUE then Writeln('Could not create file: ',Filename)
    else begin
      Result:=true; FileClose(fh); inc(CCnt);
      end;
    end;
  end;

{ ---------------------------------------------------------------- }
// read command line options
function ReadOptions : boolean;
var
  s : string;
  i : integer;
begin
  Delta:=0; CopyTs:=false; UseCurrent:=false; ModTs:=[]; sDate:=''; sTime:='';
  HasOptions:=false;
  if ParamCount>0 then begin
    Result:=true;
    for i:=1 to ParamCount do begin
      s:=ParamStr(i);
      if (s[1]='/') or (s[1]='-') then begin
        HasOptions:=true;
        delete (s,1,1);
        if ReadOptionValue(s,'b') then begin
           Delta:=-ReadNxtInt(s,' ',0);
           end
        else if ReadOptionValue(s,'f') then begin
           Delta:=ReadNxtInt(s,' ',0);
           end
        else if ReadOptionValue(s,'d') then begin
           sDate:=s;
           end
        else if ReadOptionValue(s,'t') then begin
           sTime:=s;
           end
        else if CompareOption(s,'r') then CopyTs:=true
//        else if CompareOption(s,'n') then UseCurrent:=true
        else begin
          if Pos('c',s)>0 then Include(ModTs,mtCreate);
          if Pos('w',s)>0 then Include(ModTs,mtWrite);
          if Pos('a',s)>0 then Include(ModTs,mtAccess);
          end
        end
      else FileList.Add(s);
      end;
    if ModTs=[] then Include(ModTs,mtWrite);
    if not CopyTs and (length(sDate)=0) and (length(sTime)=0) then UseCurrent:=true;
    end
  else Result:=false;
  end;

begin
  writeln(GetVersionInfo);
  FileList:=TStringList.Create;
  if ReadOptions then begin
    nc:=0; nn:=0;
    if FileList.Count>0 then begin
      with FileList do begin
        if CopyTs then begin  // option -r
          if FileExistsXL(Strings[0]) then begin
            fTs:=GetFileTimestamps(Strings[0],false);
            IncTimeStamps(fTs);
            for i:=1 to Count-1 do if CheckFile(Strings[i],nn) and SetFileTimestamps(Strings[i],fTs) then inc(nc);
            end
          else writeln('Reference file not found: ',Strings[0]);
          end
        else if UseCurrent then begin  // set to current date and time
          fTs:=GetCurrentTimestamps;
          for i:=0 to Count-1 do begin
            if CheckFile(Strings[i],nn) and HasOptions and SetFileTimestamps(Strings[i],fTs) then inc(nc);
            end;
          end
        else begin   // change timestamps as defined in options
          for i:=0 to Count-1 do if CheckFile(Strings[i],nn) then begin
            fTs:=GetFileTimestamps(Strings[i],false);
            ModifyTimestamps(fTs);
            if SetFileTimestamps(Strings[i],fTs) then inc(nc);
            end;
          end;
        if nn>0 then writeln(' -> ',nn,' file(s) created!');
        if nc>0 then writeln(' -> ',nc,' file(s) changed!');
        end
      end
    else writeln('No files specified!');
    end
  else begin
    writeln ('Calling:');
    writeln ('  WinTouch<Exe> [Optionen] <file1> [<file2> [<file3> ...]]');
    writeln ('  command line options:');
    writeln ('   -cwa  : select timestamp(s) to be modified (default = w)');
    writeln ('   -r    : copy timestamps from filename1 (default only w) to filename2,...');
    writeln ('     -b:n  : subtract n seconds from time stamp');
    writeln ('     -f:n  : add n seconds from time stamp');
    writeln ('   -d:[[YYYY]MM]DD : set to fix date, time and omitted parts remain unchanged');
    writeln ('   -t:hh[mm[ss]]   : set to fix time, date and omitted parts remain unchanged');
    writeln ('   <none of r,d,t> : set all files to current date and time');
    writeln ('   Note: not existing files will be created');
    end;
  FileList.Free;

{$IFDEF Debug}
  write ('Strike enter key to continue ...');
  readln;
{$EndIf}
end.
